package org.bitbucket.zetro.squaredino.backend;

import java.util.List;

/**
 * An abstract automaton class for updating cell boards.
 */
public abstract class AbstractAutomaton {
	/**
	 * Standard rule.
	 */
	public static final String STANDARD_RULE = GameOfLife.STANDARD_RULE;
	/**
	 * Cellular automaton.
	 */
	protected CellularAutomaton automaton;

	/**
	 * Automaton constructor using standard rules.
	 */
	public AbstractAutomaton() {
		parseRule(AbstractAutomaton.STANDARD_RULE);
	}

	/**
	 * Automaton constructor using given rules.
	 *
	 * @param rule the used rule
	 */
	public AbstractAutomaton(String rule) {
		parseRule(rule);
	}

	private void parseRule(String rule) throws IllegalArgumentException {
		if (rule == null) {
			throw new IllegalArgumentException("Invalid rule: " + rule);
		}

		if (!rule.contains(":") && rule.contains("/")) {
			int c = rule.split("/").length;
			if (c == 2) {
				automaton = new GameOfLife(rule);
			} else if (c == 3) {
				automaton = new Generations(rule);
			} else {
				throw new IllegalArgumentException("Invalid rule: " + rule);
			}
		} else {
			String name;
			String value;
			if (rule.contains(":")) {
				name = rule.substring(0, rule.indexOf(':'));
				value = rule.substring(rule.indexOf(':') + 1);
			} else {
				name = rule;
				value = null;
			}

			name = name.toLowerCase();
			if (name.equals("life")) {
				automaton = new GameOfLife(value == null ? GameOfLife.STANDARD_RULE : value);
			} else if (name.equals("brians-brain") || name.equals("briansbrain") || name.equals("brian") || name.equals("brain")) {
				automaton = new BriansBrain();
			} else if (name.equals("langtons-ant") || name.equals("langtonsant") || name.equals("langton") || name.equals("ant")) {
				automaton = new LangtonsAnt(value == null ? LangtonsAnt.STANDARD_RULE : value);
			} else if (name.equals("turmite")) {
				automaton = new Turmite(value == null ? Turmite.STANDARD_RULE : value);
			} else if (name.equals("hpp")) {
				automaton = new HPP();
			} else if (name.equals("generations")) {
				automaton = new Generations(value == null ? Generations.STANDARD_RULE : value);
			} else if (name.equals("margolus")) {
				automaton = new Margolus(value == null ? Margolus.STANDARD_RULE : value);
			} else if (name.equals("rules-table") || name.equals("rules table") || name.equals("rulestable")) {
				automaton = new RulesTable(value == null ? RulesTable.STANDARD_RULE : value);
			} else if (rule.startsWith("M")) {
				automaton = new Margolus(rule);
			} else if (Character.isDigit(rule.charAt(0))) {
				automaton = new RulesTable(rule);
			} else {
				throw new IllegalArgumentException("Unknown rule type in rule \"" + rule + "\"");
			}
		}
	}

	/**
	 * Sets the rules for this automaton.
	 *
	 * @param rule the new rule
	 * @throws IllegalArgumentException thrown if rule is null or invalid.
	 */
	public void setRule(String rule) throws IllegalArgumentException {
		parseRule(rule);
	}

	/**
	 * Returns a list of possible (or common) states.
	 *
	 * @return a list of possible (or common) states
	 */
	public List<Cell> getStates() {
		return automaton.getStates();
	}

	/**
	 * Returns the preferred color of a cell (state).
	 *
	 * @param cell the cell
	 * @return the preferred color
	 */
	public int getColor(Cell cell) {
		return automaton.getColor(cell);
	}

	/**
	 * Returns the rules as a string.
	 *
	 * @return the rules as a string.
	 */
	public String getRuleAsString() {
		String name;
		if (automaton instanceof GameOfLife) {
			name = "";
		} else if (automaton instanceof BriansBrain) {
			name = "Brians-Brain";
		} else if (automaton instanceof LangtonsAnt) {
			name = "Langtons-Ant";
		} else if (automaton instanceof Turmite) {
			name = "Turmite";
		} else if (automaton instanceof HPP) {
			name = "HPP";
		} else if (automaton instanceof Generations) {
			name = "Generations";
		} else if (automaton instanceof Margolus) {
			name = "Margolus";
		} else if (automaton instanceof RulesTable) {
			name = "Rules-Table";
		} else {
			name = "<Unknown>";
		}

		String rule = automaton.getRule();
		return name + (rule != null ? (name.isEmpty() ? "" : ":") + rule : "");
	}
}
