package org.bitbucket.zetro.squaredino.backend;

import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import org.bitbucket.zetro.squaredino.misc.Point;
import org.bitbucket.zetro.squaredino.misc.Selection;

/**
 * Board class containing all (living) cells.
 */
public abstract class AbstractBoard implements Board {
	/**
	 * Updates this Board
	 * <code>n</code> generation/step. This implementation only call on
	 * <code>step()</code> multiple times.
	 *
	 * @param n number of generasions/steps forward
	 */
	@Override
	public void step(int n) {
		for (int i = 0; i < n; i++) {
			step();
		}
	}

	/**
	 * Adds a living cells in a board to this board on a specified position.
	 *
	 * @param board the board, mustn't be null
	 * @param x position x-coordinate
	 * @param y position y-coordinate
	 */
	@Override
	public void insertBoard(Board board, int x, int y) {
		for (Iterator<Map.Entry<Point, Cell>> it = board.getLivingCells(); it.hasNext();) {
			Map.Entry<Point, Cell> entry = it.next();

			Point p = entry.getKey();
			addCell(entry.getValue(), p.x + x, p.y + y);
		}
	}

	/**
	 * Creates and return a new board containing all living cells from this
	 * board that is inside a specified selection.
	 *
	 * @param selection board selection
	 * @return a sub-board of this board.
	 */
	@Override
	public Board getSubBoard(Selection selection) {
		return getSubBoard(selection.getTopLeft().x, selection.getTopLeft().y,
				selection.getWidth(), selection.getHeight());
	}

	@Override
	public Iterator<Entry<Point, Cell>> getLivingCells(Selection selection, int ignoreLevel) {
		if (selection == null) {
			return getLivingCells();
		}

		final int x = selection.getTopLeft().x;
		final int y = selection.getTopLeft().y;
		final int width = selection.getWidth();
		final int height = selection.getHeight();
		final Iterator<Map.Entry<Point, Cell>> it = getLivingCells();

		return new Iterator<Entry<Point, Cell>>() {
			private Entry<Point, Cell> next;

			@Override
			public boolean hasNext() {
				if (next == null) {
					next = next();
				}
				return next != null;
			}

			@Override
			public Entry<Point, Cell> next() {
				if (next != null) {
					Entry<Point, Cell> tmp = next;
					next = null;
					return tmp;
				}

				while (it.hasNext()) {
					Map.Entry<Point, Cell> entry = it.next();
					Point point = entry.getKey();

					if (point.x >= x && point.y >= y
							&& point.x <= x + width && point.y <= y + height) {
						return entry;
					}
				}
				return null;
			}

			@Override
			public void remove() {
				it.remove();
			}
		};
	}

	@Override
	public Iterator<Map.Entry<Point, Cell>> iterator() {
		return getLivingCells();
	}
}
