package org.bitbucket.zetro.squaredino.backend;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;
import org.bitbucket.zetro.squaredino.misc.Point;

/**
 * Automaton implementation for BoardImpl.
 */
public class Automaton extends AbstractAutomaton {
	/**
	 * Automaton constructor using standard rules.
	 */
	public Automaton() {
		super();
	}

	/**
	 * Automaton constructor using given rules.
	 *
	 * @param rule the used rule
	 */
	public Automaton(String rule) {
		super(rule);
	}

	/**
	 * Updates a given Board one generation/step.
	 *
	 * @param board the Board of cells that will be updated
	 */
	public void step(BoardImpl board) {
		BoardImpl successor = new BoardImpl(this); // Temporary successor board
		// An "edge point" is the position of a dead cell next to a living cell.
		Collection<Point> edgePoints = new HashSet<Point>(board.getNumberOfCells());
		Queue<Point> queue = new LinkedList<Point>(); // Queue of unhandled cells

		// Add all living cells for the first pass
		for (Iterator<Map.Entry<Point, Cell>> it = board.getLivingCells(); it.hasNext();) {
			queue.add(it.next().getKey());
		}

		// Two-pass loop, first all living and then all edge points
		boolean addEdges = true;
		while (!queue.isEmpty()) {
			Point point = queue.poll();

			// Create neighbour array and add edge points
			Cell[] neighbours = new Cell[8];
			checkNeighbour(CellularAutomaton.N, point.x + 0, point.y - 1, neighbours, edgePoints, board, addEdges);
			checkNeighbour(CellularAutomaton.NE, point.x + 1, point.y - 1, neighbours, edgePoints, board, addEdges);
			checkNeighbour(CellularAutomaton.E, point.x + 1, point.y + 0, neighbours, edgePoints, board, addEdges);
			checkNeighbour(CellularAutomaton.SE, point.x + 1, point.y + 1, neighbours, edgePoints, board, addEdges);
			checkNeighbour(CellularAutomaton.S, point.x + 0, point.y + 1, neighbours, edgePoints, board, addEdges);
			checkNeighbour(CellularAutomaton.SW, point.x - 1, point.y + 1, neighbours, edgePoints, board, addEdges);
			checkNeighbour(CellularAutomaton.W, point.x - 1, point.y + 0, neighbours, edgePoints, board, addEdges);
			checkNeighbour(CellularAutomaton.NW, point.x - 1, point.y - 1, neighbours, edgePoints, board, addEdges);

			// Update cell state
			Cell currentCell = board.getCell(point.x, point.y);
			Cell newState = automaton.getNextState(currentCell, neighbours);
			if (newState != null) {
				successor.addCell(newState, point.x, point.y);
			}

			// Check if first pass is finished
			if (queue.isEmpty() && addEdges) {
				// No more living cells in queue, first pass finished
				// Add all edge points to queue for the second pass
				queue.addAll(edgePoints);
				addEdges = false;
			}
		}

		// Finish by setting the successor's content to the board.
		board.set(successor);
	}

	private void checkNeighbour(int idx, int x, int y, Cell[] neighbours, Collection<Point> edgePoints, Board board, boolean addEdges) {
		Cell cell = getCell(board, x, y);
		if (cell != null) {
			neighbours[idx] = cell;
		} else if (addEdges) { // Add dead cell as edge point
			edgePoints.add(verifyPoint(new Point(x, y)));
		}
	}

	// Edge handler
	private Cell getCell(Board board, int x, int y) {
		// TODO: Implement edge handling
		return board.getCell(x, y);
	}

	// Edge handler
	private Point verifyPoint(Point point) {
		// TODO: Implement edge handling
		return point;
	}

	public CellularAutomaton getCA() {
		return automaton;
	}
}
