package org.bitbucket.zetro.squaredino.backend;

import java.util.Iterator;
import java.util.List;
import java.util.Map;
import org.bitbucket.zetro.squaredino.misc.Point;
import org.bitbucket.zetro.squaredino.misc.Selection;

/**
 * Board interface.
 */
public interface Board extends Iterable<Map.Entry<Point, Cell>> {
	/**
	 * Value used as infinite width or height.
	 */
	public static final int INFINITE = -1;

	/**
	 * Updates this Board one generation/step.
	 */
	public void step();

	/**
	 * Updates this Board
	 * <code>n</code> generation/step.
	 *
	 * @param n number of generasions/steps forward
	 */
	public void step(int n);

	/**
	 * Adds a living cells in a board to this board on a specified position.
	 *
	 * @param board the board, mustn't be null
	 * @param x position x-coordinate
	 * @param y position y-coordinate
	 */
	public void insertBoard(Board board, int x, int y);

	/**
	 * Adds a living cell to this board.
	 *
	 * @param cell the cell, shouldn't be null
	 * @param x x-coordinate of the cell
	 * @param y y-coordinate of the cell
	 */
	public void addCell(Cell cell, int x, int y);

	/**
	 * Removes a living cell from this board, making it a dead cell.
	 *
	 * @param x x-coordinate of the cell
	 * @param y y-coordinate of the cell
	 */
	public void removeCell(int x, int y);

	/**
	 * Removes all cells from this board. The board will be empty after this
	 * call returns.
	 */
	public void clear();

	/**
	 * Removes all cells inside the specified selection.
	 *
	 * @param selection board selection
	 */
	public void clear(Selection selection);

	/**
	 * Returns a copy of this board.
	 *
	 * @return a copy of this board
	 */
	public Board copy();

	/**
	 * Sets the rules.
	 *
	 * @param rule the new rule
	 * @throws IllegalArgumentException thrown if rule is null or invalid.
	 */
	public void setRule(String rule) throws IllegalArgumentException;

	/**
	 * Returns a living cell if it exist at the given coordinate, otherwise
	 * <code>null</code> to represent a dead cell.
	 *
	 * @param x x-coordinate of the cell
	 * @param y y-coordinate of the cell
	 * @return a living cell if it exist at the given coordinate, otherwise
	 * <code>null</code> to represent a dead cell.
	 */
	public Cell getCell(int x, int y);

	/**
	 * Creates and return a new board containing all living cells from this
	 * board that is inside a specified rectangle.
	 *
	 * @param x rectangle x-coordinate
	 * @param y rectangle y-coordinate
	 * @param width rectangle width
	 * @param height rectangle height
	 * @return a sub-board of this board.
	 */
	public Board getSubBoard(int x, int y, int width, int height);

	/**
	 * Creates and return a new board containing all living cells from this
	 * board that is inside a specified selection.
	 *
	 * @param selection board selection
	 * @return a sub-board of this board.
	 */
	public Board getSubBoard(Selection selection);

	/**
	 * An iterator of all living cells and their coordinates.
	 *
	 * @return an iterator of all living cells and their coordinates.
	 */
	public Iterator<Map.Entry<Point, Cell>> getLivingCells();

	/**
	 * An iterator of all living cells inside the given selection. All living
	 * cells will be included if
	 * <code>selection</code> is
	 * <code>null</code>.
	 *
	 * @param selection the selection
	 * @param ignoreLevel level of ignorance
	 * @return an iterator of all selected living cells inside the given
	 * selection.
	 */
	public Iterator<Map.Entry<Point, Cell>> getLivingCells(Selection selection, int ignoreLevel);

	/**
	 * Returns a list of possible (or common) states.
	 *
	 * @return a list of possible (or common) states
	 */
	public List<Cell> getStates();

	/**
	 * Returns the preferred color of a cell.
	 *
	 * @param cell the cell
	 * @return the preferred color
	 */
	public int getColor(Cell cell);

	/**
	 * Returns the number of living cells in this board.
	 *
	 * @return the number of living cells in this board
	 */
	public int getNumberOfCells();

	/**
	 * Returns the width of this board.
	 *
	 * @return the width of this board
	 */
	public int getWidth();

	/**
	 * Returns the height of this board.
	 *
	 * @return the height of this board
	 */
	public int getHeight();

	/**
	 * Returns the rules as a string.
	 *
	 * @return the rules as a string.
	 */
	public String getRuleAsString();
}
