package org.bitbucket.zetro.squaredino.backend;

import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;
import org.bitbucket.zetro.squaredino.backend.experimental.QuadBoard;
import org.bitbucket.zetro.squaredino.misc.Point;
import org.bitbucket.zetro.squaredino.misc.Selection;

/**
 * Board factory.
 */
public class BoardFactory {
	public static boolean useExperimental = false;

	// Limit access
	private BoardFactory() {
	}

	/**
	 * Creates a new board.
	 *
	 * @return the created board
	 */
	public static Board createBoard() {
		if (useExperimental) {
			return new QuadBoard();
		}
		return new BoardImpl();
	}

	/**
	 * Creates a new board with specified width and height.
	 *
	 * @param width width of the board
	 * @param height height of the board
	 * @return the created board
	 */
	public static Board createBoard(int width, int height) {
		if (useExperimental) {
			System.err.println("Warning[" + BoardFactory.class.getSimpleName() + "]: createBoard(int, int) not implemented. (using createBoard())");
			return createBoard();
		}
		return new BoardImpl(width, height);
	}

	/**
	 * Returns an unmodifiable view of the specified board.
	 *
	 * @param board the source board
	 * @return the created unmodifiable board
	 * @throws UnsupportedOperationException when attempting to modify the board
	 */
	public static Board unmodifiableBoard(Board board) {
		return new UnmodifiableBoard(board);
	}

	// ~150 lines of useless restrictions :D
	private static class UnmodifiableBoard implements Board {
		private final Board board;

		private UnmodifiableBoard(Board board) {
			if (board == null) {
				throw new NullPointerException();
			}
			this.board = board;
		}

		@Override
		public void step() {
			throw new UnsupportedOperationException();
		}

		@Override
		public void step(int n) {
			throw new UnsupportedOperationException();
		}

		@Override
		public void insertBoard(Board board, int x, int y) {
			throw new UnsupportedOperationException();
		}

		@Override
		public void addCell(Cell cell, int x, int y) {
			throw new UnsupportedOperationException();
		}

		@Override
		public void removeCell(int x, int y) {
			throw new UnsupportedOperationException();
		}

		@Override
		public void clear() {
			throw new UnsupportedOperationException();
		}

		@Override
		public void clear(Selection selection) {
			throw new UnsupportedOperationException();
		}

		@Override
		public Board copy() {
			return board.copy(); // Changes to a copy shouldn't change the original
		}

		@Override
		public void setRule(String rule) throws IllegalArgumentException {
			throw new UnsupportedOperationException();
		}

		@Override
		public Cell getCell(int x, int y) {
			return board.getCell(x, y);
		}

		@Override
		public Board getSubBoard(int x, int y, int width, int height) {
			return new UnmodifiableBoard(board.getSubBoard(x, y, width, height));
		}

		@Override
		public Board getSubBoard(Selection selection) {
			return new UnmodifiableBoard(board.getSubBoard(selection));
		}

		@Override
		public Iterator<Entry<Point, Cell>> getLivingCells() {
			return unmodifiableIterator(board.getLivingCells());
		}

		@Override
		public Iterator<Entry<Point, Cell>> getLivingCells(Selection selection, int ignoreLevel) {
			return unmodifiableIterator(board.getLivingCells(selection, ignoreLevel));
		}

		@Override
		public List<Cell> getStates() {
			return Collections.unmodifiableList(board.getStates());
		}

		@Override
		public int getColor(Cell cell) {
			return board.getColor(cell);
		}

		@Override
		public int getNumberOfCells() {
			return board.getNumberOfCells();
		}

		@Override
		public int getWidth() {
			return board.getWidth();
		}

		@Override
		public int getHeight() {
			return board.getHeight();
		}

		@Override
		public String getRuleAsString() {
			return board.getRuleAsString();
		}

		@Override
		public Iterator<Entry<Point, Cell>> iterator() {
			return unmodifiableIterator(board.iterator());
		}

		private Iterator<Entry<Point, Cell>> unmodifiableIterator(final Iterator<Entry<Point, Cell>> iterator) {
			return new Iterator<Entry<Point, Cell>>() {
				Iterator<Entry<Point, Cell>> i = iterator;

				@Override
				public boolean hasNext() {
					return i.hasNext();
				}

				@Override
				public Entry<Point, Cell> next() {
					return i.next();
				}

				@Override
				public void remove() {
					throw new UnsupportedOperationException();
				}
			};
		}
	}
}
