package org.bitbucket.zetro.squaredino.backend;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import org.bitbucket.zetro.squaredino.misc.Point;
import org.bitbucket.zetro.squaredino.misc.Selection;

/**
 * Board class containing all (living) cells.
 */
public class BoardImpl extends AbstractBoard {
	/**
	 * Value used as infinite width or height.
	 */
	private Map<Point, Cell> aliveCells;
	private final Automaton auto;
	private int width, height;

	/**
	 * Board constructor. The board will have infinite width and height.
	 */
	public BoardImpl() {
		this(new Automaton());
	}

	/**
	 * Board constructor with specified automaton. The board will have infinite
	 * width and height.
	 *
	 * @param auto the automaton that will update this board
	 */
	public BoardImpl(Automaton auto) {
		this(auto, INFINITE, INFINITE);
	}

	/**
	 * Board constructor with specified width and height.
	 *
	 * @param width width of the board
	 * @param height height of the board
	 */
	public BoardImpl(int width, int height) {
		this(new Automaton(), width, height);
	}

	/**
	 * Board constructor with specified width, height and automaton.
	 *
	 * @param auto the automaton that will update this board
	 * @param width width of the board
	 * @param height height of the board
	 */
	public BoardImpl(Automaton auto, int width, int height) {
		this.auto = auto;
		this.width = width;
		this.height = height;

		aliveCells = new HashMap<Point, Cell>();
	}

	/**
	 * Updates this Board one generation/step.
	 */
	@Override
	public void step() {
		auto.step(this);
	}

	/**
	 * Sets the content of this Board to the content of another board.
	 *
	 * @param successor the other board
	 */
	protected void set(BoardImpl successor) {
		aliveCells = new HashMap<Point, Cell>(successor.aliveCells);
	}

	/**
	 * Adds a living cell to this board.
	 *
	 * @param cell the cell, shouldn't be null
	 * @param x x-coordinate of the cell
	 * @param y y-coordinate of the cell
	 */
	@Override
	public void addCell(Cell cell, int x, int y) {
		if (auto.getCA() instanceof CellValidator) {
			aliveCells.put(new Point(x, y),
					((CellValidator) auto.getCA()).validateCell(this, cell, x, y));
		} else {
			aliveCells.put(new Point(x, y), cell);
		}
	}

	/**
	 * Removes a living cell from this board, making it a dead cell.
	 *
	 * @param x x-coordinate of the cell
	 * @param y y-coordinate of the cell
	 */
	@Override
	public void removeCell(int x, int y) {
		aliveCells.remove(new Point(x, y));
	}

	/**
	 * Removes all cells from this board. The board will be empty after this
	 * call returns.
	 */
	@Override
	public void clear() {
		aliveCells.clear();
	}

	/**
	 * Removes all cells inside the specified selection.
	 *
	 * @param selection board selection
	 */
	@Override
	public void clear(Selection selection) {
		int x = selection.getTopLeft().x;
		int y = selection.getTopLeft().y;
		int width = selection.getWidth();
		int height = selection.getHeight();

		for (Iterator<Point> it = aliveCells.keySet().iterator(); it.hasNext();) {
			Point point = it.next();

			if (point.x >= x && point.y >= y
					&& point.x <= x + width && point.y <= y + height) {
				it.remove();
			}
		}
	}

	@Override
	public Board copy() {
		Board board = new BoardImpl(auto, width, height);
		board.insertBoard(this, 0, 0);

		return board;
	}

	/**
	 * Sets the rules for the automaton in this board. <br />Syntax: '/'
	 * separates two strings containing numbers 0-8. Both 3/23 and B3/S23 is
	 * allowed.
	 *
	 * @param rule the new rule
	 * @throws IllegalArgumentException thrown if rule is null, isn't separated
	 * by '/' or separated strings contain characters out of range.
	 */
	@Override
	public void setRule(String rule) throws IllegalArgumentException {
		auto.setRule(rule);
	}

	/**
	 * Returns a living cell if it exist at the given coordinate, otherwise
	 * <code>null</code> to represent a dead cell.
	 *
	 * @param x x-coordinate of the cell
	 * @param y y-coordinate of the cell
	 * @return a living cell if it exist at the given coordinate, otherwise
	 * <code>null</code> to represent a dead cell.
	 */
	@Override
	public Cell getCell(int x, int y) {
		return aliveCells.get(new Point(x, y));
	}

	/**
	 * Creates and return a new board containing all living cells from this
	 * board that is inside a specified rectangle. Note: This method will
	 * change, use copy() to ensure the same behaviour.
	 *
	 * @param x rectangle x-coordinate
	 * @param y rectangle y-coordinate
	 * @param width rectangle width
	 * @param height rectangle height
	 * @return a sub-board of this board.
	 */
	@Override
	public BoardImpl getSubBoard(int x, int y, int width, int height) {
		BoardImpl sub = new BoardImpl(auto, width, height);

		for (Map.Entry<Point, Cell> entry : aliveCells.entrySet()) {
			Point point = entry.getKey();
			Cell cell = entry.getValue();

			if (point.x >= x && point.y >= y
					&& point.x <= x + width && point.y <= y + height) {
				sub.addCell(cell, point.x, point.y);
			}
		}

		return sub;
	}

	/**
	 * An iterator of all living cells and their coordinates.
	 *
	 * @return an iterator of all living cells and their coordinates.
	 */
	@Override
	public Iterator<Map.Entry<Point, Cell>> getLivingCells() {
		return Collections.unmodifiableMap(aliveCells).entrySet().iterator();
	}

	@Override
	public List<Cell> getStates() {
		return auto.getStates();
	}

	@Override
	public int getColor(Cell cell) {
		return auto.getColor(cell);
	}

	/**
	 * Returns the number of living cells in this board.
	 *
	 * @return the number of living cells in this board
	 */
	@Override
	public int getNumberOfCells() {
		return aliveCells.size();
	}

	/**
	 * Returns the automaton in this board.
	 *
	 * @return the automaton in this board
	 */
	public Automaton getAutomaton() {
		return auto;
	}

	/**
	 * Returns the width of this board.
	 *
	 * @return the width of this board
	 */
	@Override
	public int getWidth() {
		return width;
	}

	/**
	 * Returns the height of this board.
	 *
	 * @return the height of this board
	 */
	@Override
	public int getHeight() {
		return height;
	}

	/**
	 * Returns the rules as a string in the format B3/S23.
	 *
	 * @return the rules as a string in the format B3/S23.
	 */
	@Override
	public String getRuleAsString() {
		return auto.getRuleAsString();
	}
}
