package org.bitbucket.zetro.squaredino.backend;

import java.awt.Color;
import java.util.List;

/**
 * An abstract class for implementing diffrent cellular automatons that updates
 * depending on its neighbours.
 */
public abstract class CellularAutomaton {
	/**
	 * Constants for where neighbours should be located in the neightbourhood
	 * array.
	 */
	public static final int N = 0, NE = 1, E = 2, SE = 3, S = 4, SW = 5, W = 6, NW = 7;
	private static final Color[] COLORS = {Color.LIGHT_GRAY, Color.BLACK, Color.BLUE, Color.GREEN, Color.RED, Color.ORANGE, Color.MAGENTA, Color.CYAN};

	/**
	 * Returns the amount of neighbours of a specific cell type (state).
	 *
	 * @param neighbours neighbours
	 * @param cell cell type (state)
	 * @return the amount of neighbours of a specific cell type (state)
	 */
	public static int getCount(Cell[] neighbours, Cell cell) {
		int count = 0;
		for (Cell neighbour : neighbours) {
			if (cell.equals(neighbour)) {
				count++;
			}
		}
		return count;
	}

	/**
	 * Returns the preferred color of a state.
	 *
	 * @param state the cell state
	 * @return the preferred color
	 */
	public static int getStateBasedColor(int state) {
		return COLORS[state % COLORS.length].getRGB();
	}

	/**
	 * Returns the preferred color of a cell.
	 *
	 * @param cell the cell
	 * @return the preferred color
	 */
	public int getColor(Cell cell) {
		if (cell == null) {
			return Color.WHITE.getRGB();
		}

		return getStateBasedColor(cell.state);
	}

	/**
	 * Returns the next state of a given cell.
	 *
	 * @param cell the given cell
	 * @param neighbours neighbours
	 * @return the next state of the given cell
	 */
	public abstract Cell getNextState(Cell cell, Cell... neighbours);

	/**
	 * Returns a list of possible (or common) states.
	 *
	 * @return a list of possible (or common) states
	 */
	public abstract List<Cell> getStates();

	/**
	 * Returns the rules as a string.
	 *
	 * @return the rules as a string.
	 */
	public abstract String getRule();
}
