package org.bitbucket.zetro.squaredino.backend;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

/**
 * Implementation of (Conway's) Game of Life.
 */
public class GameOfLife extends CellularAutomaton {
	/**
	 * Standard rule.
	 */
	public static final String STANDARD_RULE = "B3/S23";
	/**
	 * States.
	 */
	public static final Cell ALIVE = Cell.createCell(1), DEAD = null;
	private Collection<Integer> born, survive;

	/**
	 * GameOfLife constructor. Standard Game of Life uses rule "B3/S23".
	 *
	 * @param rule rules
	 */
	public GameOfLife(String rule) {
		setRule(rule);
	}

	@Override
	public List<Cell> getStates() {
		return Arrays.asList(ALIVE, DEAD);
	}

	@Override
	public Cell getNextState(Cell cell, Cell... neighbours) {
		int c = getCount(neighbours, ALIVE);
		if (cell == DEAD && born.contains(c)) {
			return ALIVE;
		} else if (survive.contains(c)) {
			return cell;
		}
		return DEAD;
	}

	private void setRule(String rule) throws IllegalArgumentException {
		if (rule == null || !rule.contains("/")) {
			throw new IllegalArgumentException("Invalid rule: " + rule);
		}

		String bornPart = rule.substring(0, rule.indexOf('/'));
		String survivePart = rule.substring(rule.indexOf('/') + 1);

		if ((survivePart.startsWith("B") || survivePart.startsWith("b"))
				|| (bornPart.startsWith("S") || bornPart.startsWith("s"))) {
			String tmp = bornPart;
			bornPart = survivePart;
			survivePart = tmp;
		}

		if (bornPart.startsWith("B") || bornPart.startsWith("b")) {
			bornPart = bornPart.substring(1);
		}
		if (survivePart.startsWith("S") || survivePart.startsWith("s")) {
			survivePart = survivePart.substring(1);
		}

		this.born = parseRulePart(bornPart);
		this.survive = parseRulePart(survivePart);
	}

	private Collection<Integer> parseRulePart(String part) throws IllegalArgumentException {
		Collection<Integer> ints = new ArrayList<Integer>(part.length());
		for (char c : part.toCharArray()) {
			if (c < '0' || c > '8') {
				throw new IllegalArgumentException("Rule contained invalid character '" + c + "'");
			}
			ints.add(Integer.parseInt(String.valueOf(c)));
		}
		return ints;
	}

	@Override
	public String getRule() {
		StringBuilder sb = new StringBuilder();

		sb.append("B");
		for (Integer b : born) {
			sb.append(b);
		}
		sb.append("/");
		sb.append("S");
		for (Integer s : survive) {
			sb.append(s);
		}

		return sb.toString();
	}
}
