package org.bitbucket.zetro.squaredino.backend;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

/**
 * Implementation of Generations.
 */
public class Generations extends CellularAutomaton {
	/**
	 * Standard rule.
	 */
	public static final String STANDARD_RULE = "23/3/2";
	/**
	 * States.
	 */
	public static final Cell DEAD = null,
			ALIVE = Cell.createCell(1),
			DYING = Cell.createCell(2);
	private Collection<Integer> born, survive;
	private int numberOfStates;

	/**
	 * Generations constructor.
	 *
	 * @param rule rules
	 */
	public Generations(String rule) {
		setRule(rule);
	}

	private void setRule(String rule) throws IllegalArgumentException {
		if (rule == null || !rule.contains("/")) {
			throw new IllegalArgumentException("Invalid rule: " + rule);
		}

		String[] parts = rule.split("/");
		if (parts.length != 3) {
			throw new IllegalArgumentException("Invalid rule: " + rule);
		}

		this.survive = parseRulePart(parts[0]);
		this.born = parseRulePart(parts[1]);
		if (!parts[2].isEmpty()) {
			this.numberOfStates = Integer.parseInt(parts[2]);
			if (numberOfStates < 2) {
				throw new IllegalArgumentException("Invalid rule: " + rule);
			}
		}
	}

	private Collection<Integer> parseRulePart(String part) throws IllegalArgumentException {
		Collection<Integer> ints = new ArrayList<Integer>(part.length());
		for (char c : part.toCharArray()) {
			if (c < '0' || c > '8') {
				throw new IllegalArgumentException("Rule contained invalid character '" + c + "'");
			}
			ints.add(Integer.parseInt(String.valueOf(c)));
		}
		return ints;
	}

	@Override
	public List<Cell> getStates() {
		return Arrays.asList(ALIVE, DYING, DEAD);
	}

	@Override
	public Cell getNextState(Cell cell, Cell... neighbours) {
		int c = getCount(neighbours, ALIVE);
		if (cell == null) {
			if (born.contains(c)) {
				return ALIVE;								// Born
			}
			return DEAD;									// Still dead
		} else if (cell.equals(ALIVE)) {
			if (survive.contains(c)) {
				return cell;								// Survived
			}
			if (numberOfStates != 2) {
				return DYING;								// Dying
			} else { // Life / Life-like
				return DEAD;								// Dead
			}
		} else if (cell.state >= DYING.state && cell.state < numberOfStates) {
			int state = cell.state + 1;
			if (state < numberOfStates) {
				return Cell.createCell(state);				// More dying
			} else {
				return DEAD;								// Dead
			}
		}
		System.err.println("Warning: Invalid cell " + cell);
		return DEAD;
	}

	@Override
	public String getRule() {
		StringBuilder sb = new StringBuilder();
		for (Integer s : survive) {
			sb.append(s);
		}
		sb.append("/");
		for (Integer b : born) {
			sb.append(b);
		}
		sb.append("/").append(numberOfStates);
		return sb.toString();
	}
}
