package org.bitbucket.zetro.squaredino.backend;

import java.awt.Color;
import java.util.Arrays;
import java.util.List;

/**
 * Implementation of Langton's Ant.
 */
public class LangtonsAnt extends CellularAutomaton {
	/**
	 * Standard rule.
	 */
	public static final String STANDARD_RULE = "RL";
	private final String turns;

	/**
	 * LangtonAnt constructor. Standard Langton's Ant uses rule "RL".
	 *
	 * @param turns turn rules
	 */
	public LangtonsAnt(String turns) {
		this.turns = turns;
	}

	@Override
	public int getColor(Cell cell) {
		if (cell != null && cell.state >= turns.length()) {
			return Color.RED.getRGB();
		}

		return super.getColor(cell);
	}

	@Override
	public List<Cell> getStates() {
		// Ground 1, ant on clear ground and facing N, empty
		return Arrays.asList(Cell.createCell(1),
				Cell.createCell(turns.length()),
				null);
	}

	@Override
	public Cell getNextState(Cell cell, Cell... neighbours) {
		int cellState = cell != null ? cell.state : 0;
		int cellGround = cellState;
		if (cellGround >= turns.length()) {
			cellGround = (cellGround - turns.length()) / 4;
		}
		if (cellGround < 0 || cellGround > turns.length() - 1) {
			System.err.println("Warning: Invalid ground value " + cellGround + " in cell " + cell);
			return null;
		}

		// Check if an ant moves from this cell
		boolean isAnt = cellState >= turns.length();
		if (isAnt) {
			cellGround = (cellGround + 1) % turns.length(); // Flip ground
		}

		// Check if any orthogonal neighbour is an ant moving to this cell.
		for (int i = 0; i < neighbours.length; i += 2) {
			Cell neighbour = neighbours[i];
			if (neighbour == null) {
				continue;
			}
			int state = neighbour.state;

			// Check if neighbour is an ant
			isAnt = state >= turns.length();
			if (isAnt) {
				// Retrieve what the ant stands on
				int ground = state;
				if (ground >= turns.length()) {
					ground = (ground - turns.length()) / 4;
				}
				if (cellGround < 0 || cellGround > turns.length() - 1) {
					System.err.println("Warning: Invalid ground value " + cellGround + " in cell " + neighbour);
					return null;
				}

				// Turn right or left
				int direction = (state - turns.length()) % 4;
				direction = (direction + (turns.charAt(ground) == 'R' ? 1 : -1) + 4) % 4;

				// Did the ant turn and move to towards this cell?
				if ((direction + 2) % 4 == i / 2) {
					// Calculate new state
					state = turns.length() + cellGround * 4 + direction;
					return Cell.createCell(state);
				}
			}
		}

		// Ground cell
		return Cell.createCell(cellGround);
	}

	@Override
	public String getRule() {
		return turns;
	}
}
