package org.bitbucket.zetro.squaredino.backend;

import java.awt.Color;
import java.util.ArrayList;
import java.util.List;

/**
 * Implementation of Margolus.
 */
public class Margolus extends CellularAutomaton implements CellValidator {
	/**
	 * Standard rule.
	 */
	public static final String STANDARD_RULE = "MS,D0;8;4;3;2;5;9;7;1;6;10;11;12;13;14;15";
	private final String rule;
	private final int[] transitionArray;
	private final int states;

	/**
	 * Margolus constructor.<br/>Rule syntax:
	 * <code>M&lt;number of
	 * states&gt;,D&lt;block transition #1&gt;;&lt;#2&gt;;&lt;#3&gt;;&lt;...&gt;;&lt;#(states^4)&gt;</code>
	 *
	 * @param rule rule string
	 */
	public Margolus(String rule) {
		//rule = "MS,D0;8;4;3;2;5;9;7;1;6;10;11;12;13;14;15"; // BBM
		//rule = "MS,D0;4;8;12;4;12;12;13;8;12;12;14;12;13;14;15"; // Sand
		//rule = "MU,D0;16;2;19;64;65;66;7;8;24;10;11;76;13;14;15;16;80;18;19;80;81;82;23;24;25;26;27;28;29;30;31;32;96;34;35;96;97;98;39;40;41;42;43;44;45;46;47;48;49;50;51;52;53;54;55;56;57;58;59;60;61;62;63;64;80;66;67;80;84;70;71;72;88;74;75;76;77;78;79;80;81;82;83;84;85;86;87;88;89;90;91;92;93;94;95;96;97;98;99;100;101;102;103;104;105;106;107;108;109;110;111;112;113;114;115;116;117;118;119;120;121;122;123;124;125;126;127;128;144;130;131;144;148;134;135;136;152;138;139;140;141;142;143;144;145;146;147;148;149;150;151;152;153;154;155;156;157;158;159;160;161;162;163;164;165;166;167;168;169;170;171;172;173;174;175;176;177;178;179;180;181;182;183;184;185;186;187;188;189;190;191;192;193;194;195;196;197;198;199;200;201;202;203;204;205;206;207;208;209;210;211;212;213;214;215;216;217;218;219;220;221;222;223;224;225;226;227;228;229;230;231;232;233;234;235;236;237;238;239;240;241;242;243;244;245;246;247;248;249;250;251;252;253;254;255";

		states = rule.charAt(1) - 'S' + 2;
		transitionArray = new int[states * states * states * states];

		String[] transitions = rule.substring(rule.indexOf('D') + 1).split(";");
		for (int i = 0; i < transitions.length; i++) {
			int transitionState = Integer.parseInt(transitions[i].trim());
			transitionArray[i] = transitionState;
		}

		this.rule = rule;
	}

	@Override
	public int getColor(Cell cell) {
		if (cell == null) {
			return super.getColor(cell);
		}
		// Remove blinking (almost)
		int state = (cell.state + 1) / 2;
		Color c = new Color(CellularAutomaton.getStateBasedColor(state));
		if (state == 1) {
			c = Color.DARK_GRAY;
		}
		if (isOdd(cell)) {
			c = c.darker();
		}
		return c.getRGB();
	}

	@Override
	public List<Cell> getStates() {
		List<Cell> cellStates = new ArrayList<Cell>();
		for (int i = 1; i <= states * 2; i++) {
			cellStates.add(Cell.createCell(i));
		}
		cellStates.add(null);

		return cellStates;
	}

	@Override
	public Cell getNextState(Cell cell, Cell... neighbours) {
		if (cell == null) {
			return null;
		}
		boolean odd = isOdd(cell);
		boolean change = false;

		int cellPos;
		Cell nw, ne, sw, se;
		if (odd) {
			change = true;
			cellPos = 0;
			nw = cell;
			ne = neighbours[E];
			sw = neighbours[S];
			se = neighbours[SE];
		} else {
			if (isOdd(neighbours[NW])) {
				change = true;
				cellPos = 3;
				nw = neighbours[NW];
				ne = neighbours[N];
				sw = neighbours[W];
				se = cell;
			} else if (isOdd(neighbours[N])) {
				cellPos = 2;
				nw = neighbours[N];
				ne = neighbours[NE];
				sw = cell;
				se = neighbours[E];
			} else if (isOdd(neighbours[W])) {
				cellPos = 1;
				nw = neighbours[W];
				ne = cell;
				sw = neighbours[SW];
				se = neighbours[S];
			} else {
				int newState = getSubBlockState(cell);
				change = !(isOdd(neighbours[N]) || isOdd(neighbours[E]) || isOdd(neighbours[S]) || isOdd(neighbours[W]));
				newState *= 2;
				if (change) {
					if (odd) {
						newState++;
					}
				} else if (!odd) {
					newState++;
				}
				return Cell.createCell(newState + 1);
			}
		}

		int blockState = getBlock(nw, ne, sw, se);
		if (blockState == -1) {
			return null;
		}
		int transitionState;
		if (blockState < transitionArray.length) {
			transitionState = transitionArray[blockState];
		} else {
			transitionState = blockState;
		}

		int newState = transitionState;
		for (int i = 0; i <= cellPos; i++) {
			if (i == cellPos) {
				newState %= states;
			} else {
				newState /= states;
			}
		}
		newState *= 2;
		if (change) {
			if (odd) {
				newState++;
			}
		} else if (!odd) {
			newState++;
		}

		return Cell.createCell(newState + 1);
	}

	private int subBlocksToBlock(int nw, int ne, int sw, int se) {
		return (nw + (ne + (sw + se * states) * states) * states);
	}

	private int getBlock(Cell nw, Cell ne, Cell sw, Cell se) {
		if ((isOdd(nw) ? 1 : 0) + (isOdd(ne) ? 1 : 0) + (isOdd(sw) ? 1 : 0) + (isOdd(se) ? 1 : 0) != 1) {
			System.err.println("Warning: Invalid amount of odd cells!");
			return -1;
		}
		return subBlocksToBlock(getSubBlockState(nw), getSubBlockState(ne),
				getSubBlockState(sw), getSubBlockState(se));
	}

	private int getSubBlockState(Cell cell) {
		int state = (cell != null ? (cell.state - 1) / 2 : transitionArray.length);
		return state;
	}

	private static boolean isOdd(Cell cell) {
		return (cell != null ? cell.state % 2 : 0) != 0;
	}

	@Override
	public String getRule() {
		return rule;
	}

	@Override
	public Cell validateCell(Board board, Cell cell, int x, int y) {
		boolean odd;

		// Check if any orthogonal neighbour at range 2 is odd
		Cell possibleOdd;
		if ((possibleOdd = board.getCell(x, y - 2)) != null) {
			odd = isOdd(possibleOdd);
		} else if ((possibleOdd = board.getCell(x + 2, y)) != null) {
			odd = isOdd(possibleOdd);
		} else if ((possibleOdd = board.getCell(x, y - 2)) != null) {
			odd = isOdd(possibleOdd);
		} else if ((possibleOdd = board.getCell(x - 2, y)) != null) {
			odd = isOdd(possibleOdd);
		} else {
			return cell;
		}

		// Correct cell
		int state = getSubBlockState(cell) * 2;
		if (!odd) {
			state++;
		}
		return Cell.createCell(state + 1);
	}
}
