package org.bitbucket.zetro.squaredino.backend;

import java.util.ArrayList;
import java.util.List;

/**
 * Implementation of MCell rules tables that support all totalistic cellular
 * automata.
 */
public class RulesTable extends CellularAutomaton {
	private static enum NeighbourhoodType {
		MOORE, VON_NEUMANN
	}
	/**
	 * Standard rule.
	 */
	public static final String STANDARD_RULE = "1,0,0,0,0,0,1,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0";
	private final int[][] ruleTable;
	private final int states;
	private final NeighbourhoodType neighbourhoodType;
	private final boolean oneBitFiring;

	/**
	 * RulesTable constructor.
	 *
	 * @param rule the rules table
	 */
	public RulesTable(String rule) {
		if (rule == null) {
			throw new IllegalArgumentException("Invalid rule: " + rule);
		}

		int numberOfStates = 0;
		List<Integer> numbers = new ArrayList<Integer>();
		String[] ss = rule.split(",");
		for (String s : ss) {
			int number = Integer.parseInt(s.trim());
			numbers.add(number);

			if (number + 1 > numberOfStates) {
				numberOfStates = number + 1;
			}
		}

		int type = numbers.remove(0);
		int activity = numbers.remove(0);
		int firingType = numbers.remove(0);

		this.states = numberOfStates;
		if (type == 0) {
			neighbourhoodType = NeighbourhoodType.VON_NEUMANN;
		} else if (type == 1) {
			neighbourhoodType = NeighbourhoodType.MOORE;
		} else {
			throw new IllegalArgumentException("Invalid rule: " + rule);
		}
		this.oneBitFiring = firingType != 0;

		ruleTable = new int[numberOfStates][10];
		for (int i = 0; i < numbers.size(); i++) {
			ruleTable[i / 10][i % 10] = numbers.get(i);
		}
	}

	@Override
	public List<Cell> getStates() {
		List<Cell> cellStates = new ArrayList<Cell>();
		for (int i = 1; i < states; i++) {
			cellStates.add(Cell.createCell(i));
		}
		cellStates.add(null);

		return cellStates;
	}

	@Override
	public Cell getNextState(Cell cell, Cell... neighbours) {
		int currentState = cell != null ? cell.state : 0;

		int count = 0;
		for (int i = 0; i < neighbours.length;) {
			Cell neighbour = neighbours[i];
			if (isFiring(neighbour)) {
				count++;
			}

			if (neighbourhoodType == NeighbourhoodType.MOORE) {
				i++;
			} else if (neighbourhoodType == NeighbourhoodType.VON_NEUMANN) {
				i += 2;
			}
		}

		int newState = ruleTable[currentState][count];

		return Cell.createCell(newState);
	}

	private boolean isFiring(Cell cell) {
		return cell != null ? (cell.state & (oneBitFiring ? 1 : 0xff)) == 1 : false;
	}

	@Override
	public String getRule() {
		StringBuilder sb = new StringBuilder();

		int type = neighbourhoodType == NeighbourhoodType.VON_NEUMANN ? 0 : 1;
		sb.append("1,").append(type).append(",").append(oneBitFiring ? "1" : "0").append(",");
		for (int[] stateRow : ruleTable) {
			sb.append(" ");
			for (int transition : stateRow) {
				sb.append(transition).append(",");
			}
		}

		return sb.substring(0, sb.length() - 1).toString();
	}
}
