package org.bitbucket.zetro.squaredino.backend;

import java.util.Arrays;
import java.util.List;

/**
 * Implementation of Turmite.
 */
public class Turmite extends CellularAutomaton {
	/**
	 * Standard rule.
	 */
	public static final String STANDARD_RULE = "1R1|0L1 1N0|0N0";
	private static final int GROUND_START = 8;
	private static final int STATE_START = 3;
	private final String[] table;

	/**
	 * Turmite constructor.
	 *
	 * @param turns turn rules
	 */
	public Turmite(String turns) {
		//turns = "1R0|0L0"; // Langton's Ant
		//turns = "1R0|0N0"; // Binary counter
		//turns = "1R1|0L1 1N0|0N0"; // Langton's Ant (2-state) 
		//turns = "0N0|1U0"; // Bounce
		//turns = "1L1|1L1 1R1|0N0"; // Fibonacci Spiral

		this.table = turns.split(" ");
	}

	@Override
	public List<Cell> getStates() {
		// Ground 1, ant on clear ground and facing N, empty
		return Arrays.asList(Cell.createCell(1 << GROUND_START), Cell.createCell(1), null);
	}

	@Override
	public Cell getNextState(Cell cell, Cell... neighbours) {
		int cellState = cell != null ? cell.state : 0;
		int cellGround = cell != null ? cell.state >>> GROUND_START : 0;
		if (cellGround < 0 || cellGround > table[0].length() / 4) {
			System.err.println("Warning: Invalid ground value " + cellGround + " in cell " + cell);
			return null;
		}

		// Check if an ant moves from this cell
		boolean isAnt = (cellState & 0x01) == 1;
		if (isAnt) {
			int antState = (cellState & (((1 << (GROUND_START - STATE_START)) - 1) << STATE_START)) >>> STATE_START;
			if (antState < 0 || antState > table.length - 1) {
				System.err.println("Warning: Invalid ant state " + antState + " in cell " + cell);
				return null;
			}

			// Update ground
			cellGround = table[antState].charAt(0 + cellGround * 4) - '0';
		}

		// Check if any orthogonal neighbour is an ant moving to this cell.
		for (int i = 0; i < neighbours.length; i += 2) {
			Cell neighbour = neighbours[i];
			if (neighbour == null) {
				continue;
			}
			int state = neighbour.state;

			// Check if neighbour is an ant moving to this cell.
			isAnt = (state & 0x01) == 1;
			if (isAnt) {
				// Retrieve what the ant stands on
				int ground = state >>> GROUND_START;
				if (ground < 0 || ground > table[0].length() / 4) {
					System.err.println("Warning: Invalid ground value " + ground + " in cell " + neighbour);
					return null;
				}

				// Current state of the ant
				int antState = (state & (((1 << (GROUND_START - STATE_START)) - 1) << STATE_START)) >>> STATE_START;
				if (antState < 0 || antState > table.length - 1) {
					System.err.println("Warning: Invalid ant state " + antState + " in cell " + neighbour);
					return null;
				}

				// Update direction
				int turn = table[antState].charAt(1 + ground * 4);
				int direction = (state & 0x06) >>> 1;
				if (turn == 'R') { // Right (relative)
					direction++;
				} else if (turn == 'L') { // Left (relative)
					direction--;
				} else if (turn == 'U') { // U-turn (relative)
					direction += 2;
				} else if (turn == 'n') { // North (absolute)
					direction = 0;
				} else if (turn == 'e') { // East (absolute)
					direction = 1;
				} else if (turn == 's') { // South (absolute)
					direction = 2;
				} else if (turn == 'w') { // West (absolute)
					direction = 3;
				} else if (turn != 'N') { // No turn
					System.err.println("Unknown turn: " + turn);
				}
				direction = (direction + 4) % 4;

				// Did the ant turn and move to towards this cell?
				if ((direction + 2) % 4 == i / 2) {
					// Update state
					antState = table[antState].charAt(2 + ground * 4) - '0';

					// Calculate new state
					state = (cellGround << GROUND_START) + (antState << STATE_START) + (direction << 1) + 1;
					return Cell.createCell(state);
				}
			}
		}

		// Ground cell
		return Cell.createCell(cellGround << GROUND_START);
	}

	@Override
	public String getRule() {
		StringBuilder sb = new StringBuilder();

		for (String entry : table) {
			sb.append(" ");
			sb.append(entry);
		}

		return sb.substring(1);
	}
}
