package org.bitbucket.zetro.squaredino.backend.experimental;

import java.util.ArrayList;
import java.util.Arrays;
import org.bitbucket.zetro.squaredino.backend.AbstractAutomaton;
import org.bitbucket.zetro.squaredino.backend.Cell;
import org.bitbucket.zetro.squaredino.backend.CellularAutomaton;

/**
 * Automaton implementation that uses the Hashlife algorithm.
 */
public class HashAutomaton extends AbstractAutomaton {
	private static final int NW = 0, NE = 1, SE = 2, SW = 3;
	public static boolean useLinear = false;
	public static int stepSize = 0;
	private HashQuadTree<Cell> hashTree;
	private HashNodeCache cache;
	private boolean linear;
	private int prevStepSize;

	/**
	 * HashAutomaton constructor.
	 */
	public HashAutomaton() {
		linear = useLinear;
		prevStepSize = stepSize;
	}

	/**
	 * Updates a given node one generation/step.
	 *
	 * @param node the node that will be updated
	 * @return the updated node
	 */
	public HashNode step(HashNode node) {
		if (useLinear != linear || stepSize != prevStepSize) {
			hashTree.clearCache();
			linear = useLinear;
			prevStepSize = stepSize;
		}

		// Create an expanded node with the current node in the middle
		HashNode expanded = expand(node);		// level == node.level + 1
		// Expand again in case some cells outside updates
		expanded = expand(expanded);			// level == node.level + 2

		HashNode n;
		if (useLinear || stepSize == 0) {
			n = nextGeneration(expanded, 0);	// level == node.level + 1
		} else {
			while (expanded.level < stepSize) {
				expanded = expand(expanded);
			}

			//System.out.print("Step size: " + (1 << stepSize) + ", population: " + expanded.size());
			n = nextGeneration(expanded, stepSize);		// level == node.level + 1
			//System.out.println(", new population: " + n.size());
		}
		// Compact node, usally undoes the second expand
		HashNode c = compact(n);				// level <= node.level + 1

		return c;
	}

	private HashNode nextGeneration(HashNode node, int linearLevel) {
		if (node.next != null) {
			return node.next;
		}

		HashNode nextNode;
		if (node.level == 2) {
			HashNode nw = getNextState(node.nw.se, merge(sum(node.nw, NE), sum(node.ne, NW, SW), sum(node.se, NW), sum(node.sw, NE, NW), sum(node.nw, SW, NW)));
			HashNode ne = getNextState(node.ne.sw, merge(sum(node.ne, NW, NE, SE), sum(node.se, NE, NW), sum(node.sw, NE), sum(node.nw, SE, NE)));
			HashNode sw = getNextState(node.sw.ne, merge(sum(node.nw, SE), sum(node.ne, SW), sum(node.se, NW, SW), sum(node.sw, SE, SW, NW), sum(node.nw, SW)));
			HashNode se = getNextState(node.se.nw, merge(sum(node.ne, SW, SE), sum(node.se, NE, SE, SW), sum(node.sw, SE, NE), sum(node.nw, SE)));

			nextNode = HashNode.create(cache, nw, ne, sw, se);
		} else {
			HashNode nwc, top, nec, lef, cen, rig, swc, dwn, sec;
			if (node.level - 2 > linearLevel) { // Linear
				nwc = centeredSubnode(node.nw);				// nw center
				top = centeredHorizontal(node.nw, node.ne);	// top
				nec = centeredSubnode(node.ne);				// ne center
				lef = centeredVertical(node.nw, node.sw);	// left
				cen = centeredSubSubnode(node);				// center
				rig = centeredVertical(node.ne, node.se);	// right
				swc = centeredSubnode(node.sw);				// sw center
				dwn = centeredHorizontal(node.sw, node.se);	// down
				sec = centeredSubnode(node.se);				// se center
			} else { // Exponential
				nwc = nextGeneration(node.nw, linearLevel);				// nw center
				top = horizontalForward(node.nw, node.ne, linearLevel);	// top
				nec = nextGeneration(node.ne, linearLevel);				// ne center
				lef = verticalForward(node.nw, node.sw, linearLevel);	// left
				cen = centeredForward(node, linearLevel);				// center
				rig = verticalForward(node.ne, node.se, linearLevel);	// right
				swc = nextGeneration(node.sw, linearLevel);				// sw center
				dwn = horizontalForward(node.sw, node.se, linearLevel);	// down
				sec = nextGeneration(node.se, linearLevel);				// se center
			}

			nextNode = HashNode.create(cache,
					nextGeneration(HashNode.create(cache, nwc, top, lef, cen), linearLevel),
					nextGeneration(HashNode.create(cache, top, nec, cen, rig), linearLevel),
					nextGeneration(HashNode.create(cache, lef, cen, swc, dwn), linearLevel),
					nextGeneration(HashNode.create(cache, cen, rig, dwn, sec), linearLevel));
		}

		node.next = nextNode;
		return node.next;
	}

	private HashNode horizontalForward(HashNode w, HashNode e, int linearLevel) {
		return nextGeneration(HashNode.create(cache, w.ne, e.nw, w.se, e.sw), linearLevel);
	}

	private HashNode verticalForward(HashNode n, HashNode s, int linearLevel) {
		return nextGeneration(HashNode.create(cache, n.sw, n.se, s.nw, s.ne), linearLevel);
	}

	private HashNode centeredForward(HashNode c, int linearLevel) {
		return nextGeneration(HashNode.create(cache, c.nw.se, c.ne.sw, c.sw.ne, c.se.nw), linearLevel);
	}

	private HashNode centeredSubnode(HashNode c) {
		return HashNode.create(cache, c.nw.se, c.ne.sw, c.sw.ne, c.se.nw);
	}

	private HashNode centeredHorizontal(HashNode w, HashNode e) {
		return HashNode.create(cache, w.ne.se, e.nw.sw, w.se.ne, e.sw.nw);
	}

	private HashNode centeredVertical(HashNode n, HashNode s) {
		return HashNode.create(cache, n.sw.se, n.se.sw, s.nw.ne, s.ne.nw);
	}

	private HashNode centeredSubSubnode(HashNode c) {
		return HashNode.create(cache, c.nw.se.se, c.ne.sw.sw, c.sw.ne.ne,
				c.se.nw.nw);
	}

	private HashNode getNextState(HashNode node, Cell... count) {
		return HashContentNode.create(cache, automaton.getNextState(getCell(node), count));
	}

	private Cell getCell(HashNode node) {
		if (node == null) {
			return null;
		} else {
			if (node instanceof HashContentNode == false) {
				throw new AssertionError(node);
			}
			@SuppressWarnings("unchecked")
			HashContentNode<Cell> current = ((HashContentNode<Cell>) node);
			return current.content;
		}
	}

	private Cell[] sum(HashNode node, int... cells) {
		Cell[] count = new Cell[cells.length];

		for (int i = 0; i < cells.length; i++) {
			int cellIndex = cells[i];
			HashNode cellNode = getNodeByIndex(node, cellIndex);
			count[i] = getCell(cellNode);
		}

		return count;
	}

	private Cell[] merge(Cell[]... parts) {
		ArrayList<Cell> all = new ArrayList<Cell>();
		for (Cell[] cells : parts) {
			all.addAll(Arrays.asList(cells));
		}
		return all.toArray(new Cell[8]);
	}

	private HashNode getNodeByIndex(HashNode node, int index) {
		if (index == NW) {
			return node.nw;
		} else if (index == NE) {
			return node.ne;
		} else if (index == SE) {
			return node.se;
		} else if (index == SW) {
			return node.sw;
		} else {
			throw new IndexOutOfBoundsException("Index: " + index);
		}
	}

	private HashNode compact(HashNode node) {
		if (node.level == 1) {
			return node;
		}
		HashNode cur = node;
		HashNode space = hashTree.getSpaceNode(cur.level - 2);
		while (cur.level >= 2
				&& cur.nw.nw == space && cur.nw.ne == space && cur.nw.sw == space
				&& cur.ne.nw == space && cur.ne.ne == space && cur.ne.se == space
				&& cur.sw.nw == space && cur.sw.sw == space && cur.sw.se == space
				&& cur.se.ne == space && cur.se.sw == space && cur.se.se == space) {
			cur = HashNode.create(cache,
					cur.nw.se, cur.ne.sw, cur.sw.ne, cur.se.nw);
			if (cur.level >= 2) {
				space = hashTree.getSpaceNode(cur.level - 2);
			}
		}
		return cur;
	}

	private HashNode expand(HashNode node) {
		HashNode space = hashTree.getSpaceNode(node.level - 1);
		HashNode nw = HashNode.create(cache, space, space, space, node.nw);
		HashNode ne = HashNode.create(cache, space, space, node.ne, space);
		HashNode sw = HashNode.create(cache, space, node.sw, space, space);
		HashNode se = HashNode.create(cache, node.se, space, space, space);
		return HashNode.create(cache, nw, ne, sw, se);
	}

	public void setHashTree(HashQuadTree<Cell> hashTree) {
		this.hashTree = hashTree;
		if (hashTree != null) {
			this.cache = hashTree.getCache();
		} else {
			cache = null;
		}
	}

	public CellularAutomaton getCA() {
		return automaton;
	}
}
