package org.bitbucket.zetro.squaredino.backend.experimental;

/**
 * An immutable quadnode. hashcode() only depends on its (immutable) child
 * nodes.
 */
public class HashNode {
	/**
	 * Node level.
	 */
	public final int level;
	/**
	 * Child nodes.
	 */
	public final HashNode nw, ne, sw, se;
	/**
	 * Next generation.
	 */
	public HashNode next;
	private int cachedSize = -1;

	/**
	 * HashNode constructor.
	 *
	 * @param nw north-west node
	 * @param ne north-east node
	 * @param sw south-west node
	 * @param se south-east node
	 */
	HashNode(HashNode nw, HashNode ne, HashNode sw, HashNode se) {
		this.level = nw != null ? nw.level + 1 : 0;
		this.nw = nw;
		this.ne = ne;
		this.sw = sw;
		this.se = se;
	}

	/**
	 * Creates a new HashNode.
	 *
	 * @param cache node cache
	 * @param nw north-west node
	 * @param ne north-east node
	 * @param sw south-west node
	 * @param se south-east node
	 * @return the new HashNode
	 */
	public static HashNode create(HashNodeCache cache, HashNode nw, HashNode ne, HashNode sw, HashNode se) {
		return cache.findOrCacheNode(new HashNode(nw, ne, sw, se));
	}

	private int hash() {
		return super.hashCode();
	}

	@Override
	public int hashCode() {
		int hash = 5;
		hash = 53 * hash + this.nw.hash();
		hash = 53 * hash + this.ne.hash();
		hash = 53 * hash + this.sw.hash();
		hash = 53 * hash + this.se.hash();
		return hash;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final HashNode other = (HashNode) obj;
		//assert this.nw.equals(other.nw) ? this.nw == other.nw : true;
		//assert this.ne.equals(other.ne) ? this.ne == other.ne : true;
		//assert this.sw.equals(other.sw) ? this.sw == other.sw : true;
		//assert this.se.equals(other.se) ? this.se == other.se : true;
		if (this.nw != other.nw) { // Works because of canonicalization
			return false;
		}
		if (this.ne != other.ne) {
			return false;
		}
		if (this.sw != other.sw) {
			return false;
		}
		if (this.se != other.se) {
			return false;
		}
		return true;
	}

	/**
	 * Returns the number of elements in this node.
	 *
	 * @return the number of elements in this node
	 */
	public int size() {
		if (cachedSize == -1) {
			cachedSize = this.nw.size() + this.ne.size() + this.sw.size() + this.se.size();
		}
		return cachedSize;
	}

	@Override
	public String toString() {
		return "HashNode{" + "nw=" + nw + ", ne=" + ne + ", sw=" + sw + ", se=" + se + '}';
	}
}
