package org.bitbucket.zetro.squaredino.backend.experimental;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Stack;
import org.bitbucket.zetro.squaredino.misc.Point;
import org.bitbucket.zetro.squaredino.misc.Selection;

/**
 * Hashed quadtree with immutable and cached nodes.
 *
 * @param <T> data type stored
 */
public class HashQuadTree<T> implements Iterable<Map.Entry<Point, T>> {
	private HashNode root;
	private HashNodeCache cache;
	private HashMap<Integer, HashNode> spaceCache;
	private Point pos;
	private int numberOfElements;

	private HashQuadTree() {
		root = null;
		cache = new HashNodeCache();
		spaceCache = new HashMap<Integer, HashNode>();
	}

	/**
	 * Creates a hashed quadtree from a normal quadtree.
	 *
	 * @param <T> data type stored
	 * @param tree the normal quadtree
	 * @return a hashed quadtree with the same content as the normal one
	 */
	public static <T> HashQuadTree<T> create(QuadTree<T> tree) {
		HashQuadTree<T> hashTree = new HashQuadTree<T>();
		hashTree.pos = tree.getPosition();

		QuadTree.Node<T> node = tree.getRoot();
		if (!node.isLeaf()) {
			HashNode newRoot = buildNode(hashTree, node, tree.getLevel());
			hashTree.root = newRoot;
		}
		if (hashTree.root != null) {
			hashTree.pos = new Point(hashTree.pos.x + (1 << hashTree.root.level - 1), hashTree.pos.y + (1 << hashTree.root.level - 1));
		}

		return hashTree;
	}

	private static <T> HashNode buildNode(HashQuadTree<T> hashTree, QuadTree.Node<T> node, int level) {
		if (node == null) {
			return hashTree.getSpaceNode(level);
		} else if (!node.isLeaf()) {
			assert level > 0;

			HashNode nw = buildNode(hashTree, node.nw, level - 1);
			HashNode ne = buildNode(hashTree, node.ne, level - 1);
			HashNode sw = buildNode(hashTree, node.sw, level - 1);
			HashNode se = buildNode(hashTree, node.se, level - 1);

			return HashNode.create(hashTree.cache, nw, ne, sw, se);
		} else if (node instanceof QuadTree.ContentNode) {
			assert level == 0;
			T content = ((QuadTree.ContentNode<T>) node).content;
			hashTree.numberOfElements++;
			return HashContentNode.create(hashTree.cache, content);
		} else {
			System.err.println("Node(" + level + "): " + node);
			throw new UnsupportedOperationException("Not yet implemented");
		}
	}

	/**
	 * Returns an node with size
	 * <code>0</code> and level
	 * <code>level</code>.
	 *
	 * @param level level of the node
	 * @return an node with size
	 * <code>0</code> and level
	 * <code>level</code>
	 */
	public HashNode getSpaceNode(int level) {
		assert level >= 0;
		HashNode cached = spaceCache.get(level);
		if (cached != null) {
			return cached;
		} else if (level == 0) {
			return HashContentNode.create(cache, null);
		}

		HashNode part = getSpaceNode(level - 1);
		HashNode space = HashNode.create(cache, part, part, part, part);
		spaceCache.put(level, space);
		return space;
	}

	/**
	 * Returns an iterator that only iterate all living cells inside the given
	 * selection. All living cells will be included if
	 * <code>selection</code> is
	 * <code>null</code>.
	 *
	 * @param selection the selection
	 * @param ignoreLevel level of ignorance
	 * @return an iterator of all living cells inside the selection
	 */
	public Iterator<Entry<Point, T>> getIterator(Selection selection, int ignoreLevel) {
		if (root == null) {
			return new HashQuadTreeIterator<T>(selection, root, pos, 0, ignoreLevel);
		}

		return new HashQuadTreeIterator<T>(selection, root, new Point(pos.x - (1 << root.level - 1), pos.y - (1 << root.level - 1)), root.level, ignoreLevel);
	}

	@Override
	public Iterator<Entry<Point, T>> iterator() {
		return getIterator(null, 0);
	}

	/**
	 * Returns the number of elements in this tree.
	 *
	 * @return the number of elements in this tree
	 */
	public int size() {
		return root != null ? root.size() : 0;
	}

	/**
	 * Removes all elements from this tree. The tree will be empty after this
	 * call.
	 */
	public void clear() {
		root = getSpaceNode(1);
		pos = new Point(0, 0);
		cache.clear();
	}

	/**
	 * Removes all nodes from the cache an reinsert all nodes in this tree. All
	 * <code>node.next</code> is set to
	 * <code>null</code>.
	 */
	public void clearCache() {
		spaceCache.clear();
		cache.clear();

		Stack<HashNode> stack = new Stack<HashNode>();
		stack.add(root);
		while (!stack.isEmpty()) {
			final HashNode node = stack.pop();

			if (node != null) {
				if (!cache.isCached(node)) {
					cache.findOrCacheNode(node);
					node.next = null;

					stack.add(node.nw);
					stack.add(node.ne);
					stack.add(node.sw);
					stack.add(node.se);
				}
			}
		}
	}

	/**
	 * Cache garbage collector. Removes all nodes from the cache an reinsert all
	 * nodes in this tree.
	 */
	public void gc() {
		System.out.print("Running garbage collector... ");
		int prevSize = cache.size();
		cache.clear();
		for (HashNode space : spaceCache.values()) {
			cache.findOrCacheNode(space);
		}

		Stack<HashNode> stack = new Stack<HashNode>();
		stack.add(root);
		while (!stack.isEmpty()) {
			final HashNode node = stack.pop();

			if (node != null) {
				if (!cache.isCached(node)) {
					cache.findOrCacheNode(node);
					stack.add(node.next);
					stack.add(node.nw);
					stack.add(node.ne);
					stack.add(node.sw);
					stack.add(node.se);
				}
			}
		}
		System.out.println((prevSize - cache.size()) + " of " + prevSize + " nodes removed from cache.");
	}

	/**
	 * Returns the node cache used by this tree.
	 *
	 * @return the node cache used by this tree
	 */
	public HashNodeCache getCache() {
		return cache;
	}

	void setRoot(HashNode root) {
		this.root = root;
	}

	HashNode getRoot() {
		return root;
	}

	@Override
	public String toString() {
		return "HashQuadTree{" + "root=" + root + '}';
	}
}
