package org.bitbucket.zetro.squaredino.backend.experimental;

import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Stack;
import org.bitbucket.zetro.squaredino.misc.Point;
import org.bitbucket.zetro.squaredino.misc.Selection;

/**
 * An iterator implementation for QuadTree.
 *
 * @param <T> data type stored in the quadtree
 */
public class HashQuadTreeIterator<T> implements Iterator<Entry<Point, T>> {
	private final Selection selection;
	private Stack<HashNode> stack;
	private Stack<Point> positions;
	private Entry<Point, T> next;
	private final int ignoreLevel;

	/**
	 * QuadTreeIterator constructor.
	 *
	 * @param node iterated node
	 * @param position node position
	 */
	HashQuadTreeIterator(Selection selection, HashNode node, Point position, int levels, int ignoreLevel) {
		this.selection = selection;
		this.ignoreLevel = ignoreLevel;

		if (node != null) {
			stack = new Stack<HashNode>();
			positions = new Stack<Point>();

			add(node, position);
		}
	}

	private void add(HashNode node, Point pos) {
		if (node == null) {
			return;
		}

		if (node.level == 0 || node.size() != 0) {
			if (selection != null) {
				if (pos.x + (1 << node.level) - 1 < selection.getTopLeft().x) {
					return;
				}
				if (pos.y + (1 << node.level) - 1 < selection.getTopLeft().y) {
					return;
				}
				if (pos.x > selection.getBottomRight().x) {
					return;
				}
				if (pos.y > selection.getBottomRight().y) {
					return;
				}
			}

			stack.add(node);
			positions.add(pos);
		}
	}

	@Override
	public boolean hasNext() {
		if (stack != null && next == null) {
			next = next();
		}
		return next != null;
	}

	@Override
	public Entry<Point, T> next() {
		if (next != null) {
			Entry<Point, T> tmp = next;
			next = null;
			return tmp;
		}

		while (!stack.isEmpty()) {
			final HashNode node = stack.pop();
			final Point pos = positions.pop();
			if (node instanceof HashContentNode || (node.level == ignoreLevel && node.size() != 0)) {
				if (node instanceof HashContentNode) {
					@SuppressWarnings("unchecked")
					HashContentNode<T> contentNode = ((HashContentNode<T>) node);
					if (contentNode.content == null) {
						continue;
					}
				}
				return new Map.Entry<Point, T>() {
					@Override
					public Point getKey() {
						return pos;
					}

					@Override
					@SuppressWarnings("unchecked")
					public T getValue() {
						if (node instanceof HashContentNode) {
							return (T) ((HashContentNode) node).content;
						}
						return null;
					}

					@Override
					public T setValue(T value) {
						throw new UnsupportedOperationException("Not supported yet.");
					}
				};
			} else if (node.level == 0) {
				continue;
			}

			int size = 1 << node.level - 1;
			add(node.se, new Point(pos.x + size, pos.y + size));
			add(node.sw, new Point(pos.x, pos.y + size));
			add(node.ne, new Point(pos.x + size, pos.y));
			add(node.nw, new Point(pos.x, pos.y));
		}
		return null;
	}

	@Override
	public void remove() {
		throw new UnsupportedOperationException("Not supported yet.");
	}
}
