package org.bitbucket.zetro.squaredino.backend.experimental;

import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import org.bitbucket.zetro.squaredino.backend.AbstractBoard;
import org.bitbucket.zetro.squaredino.backend.Board;
import org.bitbucket.zetro.squaredino.backend.Cell;
import org.bitbucket.zetro.squaredino.backend.CellValidator;
import org.bitbucket.zetro.squaredino.backend.CellularAutomaton;
import org.bitbucket.zetro.squaredino.misc.Point;
import org.bitbucket.zetro.squaredino.misc.Selection;

/**
 * An experimental board class using quadtrees.
 */
public class QuadBoard extends AbstractBoard {
	private static final int CACHE_SIZE = 5000000;
	private QuadTree<Cell> tree;
	private HashQuadTree<Cell> hashTree;
	private HashAutomaton auto;

	/**
	 * QuadBoard constructor.
	 */
	public QuadBoard() {
		tree = new QuadTree<Cell>();
		auto = new HashAutomaton();
	}

	private void treeChanged() {
		if (hashTree != null) {
			rebuildQuadTree();
		}
	}

	private void buildHashTree() {
		System.out.print("Building hashtree... ");

		hashTree = HashQuadTree.create(tree);
		auto.setHashTree(hashTree);
		tree = null;

		System.out.println("done.");
	}

	private void rebuildQuadTree() {
		System.out.print("Rebuilding quadtree... ");

		tree = new QuadTree<Cell>();
		for (Map.Entry<Point, Cell> en : hashTree) {
			Point p = en.getKey();
			Cell cell = en.getValue();

			tree.add(p, cell);
		}
		hashTree = null;
		auto.setHashTree(null);

		System.out.println("done.");
	}

	@Override
	public void step() {
		if (hashTree == null) {
			buildHashTree();
		} else if (tree != null) {
			throw new AssertionError();
		}

		HashNode cur = hashTree.getRoot();
		if (cur != null) {
			HashNode next = auto.step(cur);
			hashTree.setRoot(next);
			if (hashTree.getCache().size() > CACHE_SIZE) {
				hashTree.gc();
			}
		}
	}

	@Override
	public void addCell(Cell cell, int x, int y) {
		treeChanged();

		if (auto.getCA() instanceof CellValidator) {
			tree.add(new Point(x, y),
					((CellValidator) auto.getCA()).validateCell(this, cell, x, y));
		} else {
			tree.add(new Point(x, y), cell);
		}
	}

	@Override
	public void removeCell(int x, int y) {
		treeChanged();

		tree.remove(new Point(x, y));
	}

	@Override
	public void clear() {
		if (hashTree != null) {
			hashTree.clear();
		} else {
			tree.clear();
		}
	}

	@Override
	public void clear(Selection selection) {
		treeChanged();

		int x = selection.getTopLeft().x;
		int y = selection.getTopLeft().y;
		int width = selection.getWidth();
		int height = selection.getHeight();

		for (Iterator<Map.Entry<Point, Cell>> it = getLivingCells(); it.hasNext();) {
			Map.Entry<Point, Cell> entry = it.next();
			Point point = entry.getKey();

			if (point.x >= x && point.y >= y
					&& point.x <= x + width && point.y <= y + height) {
				removeCell(point.x, point.y);
			}
		}
	}

	@Override
	public Board copy() {
		Board board = new QuadBoard();
		board.insertBoard(this, 0, 0);

		return board;
	}

	@Override
	public void setRule(String rule) throws IllegalArgumentException {
		if (rule == null || rule.isEmpty()) {
			throw new IllegalArgumentException();
		}

		auto.setRule(rule);
		if (hashTree != null) {
			hashTree.clearCache();
		}
	}

	@Override
	public Cell getCell(int x, int y) {
		treeChanged();

		return tree.get(new Point(x, y));
	}

	@Override
	public Iterator<Entry<Point, Cell>> getLivingCells() {
		return getLivingCells(null, 0);
	}

	@Override
	public Iterator<Entry<Point, Cell>> getLivingCells(Selection selection, int ignoreLevel) {
		if (hashTree != null) {
			return hashTree.getIterator(selection, ignoreLevel);
		}
		return tree.getIterator(selection, ignoreLevel);
	}

	@Override
	public Board getSubBoard(int x, int y, int width, int height) {
		QuadBoard sub = new QuadBoard();

		for (Map.Entry<Point, Cell> entry : this) {
			Point point = entry.getKey();
			Cell cell = entry.getValue();

			if (point.x >= x && point.y >= y
					&& point.x <= x + width && point.y <= y + height) {
				sub.addCell(cell, point.x, point.y);
			}
		}

		return sub;
	}

	@Override
	public List<Cell> getStates() {
		return Collections.unmodifiableList(auto.getStates());
	}

	@Override
	public int getColor(Cell cell) {
		if (auto != null) {
			return auto.getColor(cell);
		}
		return CellularAutomaton.getStateBasedColor(cell != null ? cell.state : 0);
	}

	@Override
	public int getNumberOfCells() {
		if (hashTree != null) {
			return hashTree.size();
		}
		return tree.size();
	}

	@Override
	public int getWidth() {
		//	throw new UnsupportedOperationException("Not supported yet.");
		return Board.INFINITE;
	}

	@Override
	public int getHeight() {
		//	throw new UnsupportedOperationException("Not supported yet.");
		return Board.INFINITE;
	}

	@Override
	public String getRuleAsString() {
		return auto.getRuleAsString();
	}
}
