package org.bitbucket.zetro.squaredino.backend.experimental;

import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Stack;
import org.bitbucket.zetro.squaredino.misc.Point;
import org.bitbucket.zetro.squaredino.misc.Selection;

/**
 * A point quadtree, data is added and accessed using a Point.
 *
 * @param <T> stored data type
 */
public class QuadTree<T> implements Iterable<Map.Entry<Point, T>> {
	static class Node<T> {
		public Node<T> nw, ne, sw, se;

		private Node() {
			this(null, null, null, null);
		}

		private Node(Node<T> nw, Node<T> ne, Node<T> sw, Node<T> se) {
			this.nw = nw;
			this.ne = ne;
			this.sw = sw;
			this.se = se;
		}

		@Override
		public String toString() {
			return "Node{" + "nw=" + nw + ", ne=" + ne + ", sw=" + sw + ", se=" + se + '}';
		}

		public boolean isLeaf() {
			return nw == null && ne == null && sw == null && se == null;
		}
	}

	static class ContentNode<T> extends Node<T> {
		public final T content;

		private ContentNode(T content) {
			super();

			this.content = content;
		}

		@Override
		public String toString() {
			return "ContentNode{" + "content=" + content + '}';
		}
	}
	private Node<T> root;
	private Point pos;
	private int treeSize;
	private int levels;
	private int numberOfElements;

	/**
	 * QuadTree constructor.
	 */
	public QuadTree() {
		root = new Node<T>();
		pos = new Point(0, 0);
		treeSize = 2;
		levels = 1;
	}

	private void checkSize(Point p) {
		while (p.x < pos.x || p.y < pos.y || p.x >= pos.x + treeSize || p.y >= pos.y + treeSize) {
			boolean right = p.x >= pos.x + treeSize;
			boolean down = p.y >= pos.y + treeSize;

			expand(right, down);
		}
	}

	private void expand(boolean right, boolean down) {
		Node<T> old = root;
		root = new Node<T>();

		int x = pos.x, y = pos.y;
		if (!right) {
			x -= treeSize;
		}
		if (!down) {
			y -= treeSize;
		}
		if (!old.isLeaf()) {
			setNode(root, old, !right, !down);
		}
		pos = new Point(x, y);

		treeSize *= 2;
		levels++;
	}

	/**
	 * Adds an element at the posision given by
	 * <code>p</code>.
	 *
	 * @param p the point index
	 * @param value the value
	 */
	public void add(Point p, T value) {
		checkSize(p);

		Node<T> n = root;
		int macro = treeSize;
		int x = pos.x, y = pos.y;
		while (macro >= 1) {
			if (p.x < x || p.y < y || p.x >= x + macro || p.y >= y + macro) {
				System.err.printf("macro=%d, x=%d, y=%d, px=%d, py=%d\n", macro, x, y, p.x, p.y);
				throw new AssertionError();
			}

			boolean right = p.x - x >= macro / 2;
			boolean down = p.y - y >= macro / 2;
			if (right) {
				x += macro / 2;
			}
			if (down) {
				y += macro / 2;
			}

			if (macro == 2) {
				numberOfElements++;
				Node<T> content = new ContentNode<T>(value);
				setNode(n, content, right, down);
				return;
			} else {
				Node<T> child = getNode(n, right, down);
				if (child == null) {
					child = new Node<T>();
					setNode(n, child, right, down);
				}

				n = child;
				macro /= 2;
			}
		}
		throw new AssertionError();
	}

	/**
	 * Removes the element at the posision given by
	 * <code>p</code>.
	 *
	 * @param p the point index
	 */
	public void remove(Point p) {
		Stack<Node<T>> parentNodes = new Stack<Node<T>>();
		Stack<Boolean> parentRight = new Stack<Boolean>();
		Stack<Boolean> parentDown = new Stack<Boolean>();

		Node<T> n = root;
		int macro = treeSize;
		int x = pos.x, y = pos.y;

		if (p.x < x || p.y < y || p.x >= x + macro || p.y >= y + macro) {
			return;
		}
		while (macro >= 2) {
			if (p.x < x || p.y < y || p.x >= x + macro || p.y >= y + macro) {
				System.err.printf("macro=%d, x=%d, y=%d, px=%d, py=%d\n", macro, x, y, p.x, p.y);
				throw new AssertionError();
			}

			boolean right = p.x - x >= macro / 2;
			boolean down = p.y - y >= macro / 2;
			if (right) {
				x += macro / 2;
			}
			if (down) {
				y += macro / 2;
			}

			if (macro == 2) {
				do {
					setNode(n, null, right, down);
					if (n.isLeaf() && !parentNodes.isEmpty()) {
						n = parentNodes.pop();
						right = parentRight.pop();
						down = parentDown.pop();
						continue;
					}
					break;
				} while (true);
				return;
			} else {
				parentNodes.add(n);
				parentRight.add(right);
				parentDown.add(down);
				Node<T> child = getNode(n, right, down);
				if (child == null) {
					child = new Node<T>();
					setNode(n, child, right, down);
				}

				n = child;
				macro /= 2;
			}
		}
		throw new AssertionError();
	}

	/**
	 * Returns the element at the posision given by
	 * <code>p</code>.
	 *
	 * @param p the point index
	 * @return the element on the posision given by
	 * <code>p</code>
	 */
	public T get(Point p) {
		Node<T> n = root;
		int macro = treeSize;
		int x = pos.x, y = pos.y;

		if (p.x < x || p.y < y || p.x >= x + macro || p.y >= y + macro) {
			return null;
		}
		while (macro >= 2) {
			if (p.x < x || p.y < y || p.x >= x + macro || p.y >= y + macro) {
				System.err.printf("macro=%d, x=%d, y=%d, px=%d, py=%d\n", macro, x, y, p.x, p.y);
				throw new AssertionError();
			}

			boolean right = p.x - x >= macro / 2;
			boolean down = p.y - y >= macro / 2;
			if (right) {
				x += macro / 2;
			}
			if (down) {
				y += macro / 2;
			}

			if (macro == 2) {
				Node<T> content = getNode(n, right, down);
				if (content != null) {
					return ((ContentNode<T>) content).content;
				}
				return null;
			} else {
				Node<T> child = getNode(n, right, down);
				if (child == null) {
					return null;
				}

				n = child;
				macro /= 2;
			}
		}
		throw new AssertionError();
	}

	private void setNode(Node<T> node, Node<T> value, boolean right, boolean down) {
		if (!right && !down) {
			node.nw = value;
		} else if (right && !down) {
			node.ne = value;
		} else if (!right && down) {
			node.sw = value;
		} else if (right && down) {
			node.se = value;
		} else {
			throw new AssertionError();
		}
	}

	private Node<T> getNode(Node<T> node, boolean right, boolean down) {
		if (!right && !down) {
			return node.nw;
		} else if (right && !down) {
			return node.ne;
		} else if (!right && down) {
			return node.sw;
		} else if (right && down) {
			return node.se;
		} else {
			throw new AssertionError();
		}
	}

	/**
	 * Returns an iterator that only iterate all living cells inside the given
	 * selection. All living cells will be included if
	 * <code>selection</code> is
	 * <code>null</code>.
	 *
	 * @param selection the selection
	 * @param ignoreLevel level of ignorance
	 * @return an iterator of all living cells inside the selection
	 */
	public Iterator<Entry<Point, T>> getIterator(Selection selection, int ignoreLevel) {
		return new QuadTreeIterator<T>(selection, root, pos, levels, ignoreLevel);
	}

	@Override
	public Iterator<Entry<Point, T>> iterator() {
		return getIterator(null, 0);
	}

	/**
	 * Removes all elements from this tree. The tree will be empty after this
	 * call.
	 */
	public void clear() {
		root = new Node<T>();
		pos = new Point(0, 0);
		treeSize = 2;
		levels = 1;
	}

	Node<T> getRoot() {
		return root;
	}

	Point getPosition() {
		return pos;
	}

	int getLevel() {
		return levels;
	}

	/**
	 * Returns the number of elements in this tree.
	 *
	 * @return the number of elements in this tree
	 */
	public int size() {
		return numberOfElements;
	}

	@Override
	public String toString() {
		return "QuadTree{" + "root=" + root + ", pos=" + pos + ", size=" + treeSize + '}';
	}
}
