package org.bitbucket.zetro.squaredino.backend.experimental;

import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Stack;
import org.bitbucket.zetro.squaredino.misc.Point;
import org.bitbucket.zetro.squaredino.misc.Selection;

/**
 * An iterator implementation for QuadTree.
 *
 * @param <T> data type stored in the quadtree
 */
public class QuadTreeIterator<T> implements Iterator<Entry<Point, T>> {
	private final Selection selection;
	private Stack<QuadTree.Node<T>> stack;
	private Stack<Point> positions;
	private Stack<Integer> sizes;
	private Entry<Point, T> next;
	private final int ignoreLevel;

	/**
	 * QuadTreeIterator constructor.
	 *
	 * @param node iterated node
	 * @param position node position
	 * @param levels the maximum depth in the node
	 */
	QuadTreeIterator(Selection selection, QuadTree.Node<T> node, Point position, int levels, int ignoreLevel) {
		this.selection = selection;
		this.ignoreLevel = ignoreLevel;

		stack = new Stack<QuadTree.Node<T>>();
		positions = new Stack<Point>();
		sizes = new Stack<Integer>();

		add(node, position, 1 << levels);
	}

	private void add(QuadTree.Node<T> node, Point pos, int size) {
		if (node != null) {
			if (selection != null) {
				if (pos.x + size - 1 < selection.getTopLeft().x) {
					return;
				}
				if (pos.y + size - 1 < selection.getTopLeft().y) {
					return;
				}
				if (pos.x > selection.getBottomRight().x) {
					return;
				}
				if (pos.y > selection.getBottomRight().y) {
					return;
				}
			}

			stack.add(node);
			positions.add(pos);
			sizes.add(size);
		}
	}

	@Override
	public boolean hasNext() {
		if (next == null) {
			next = next();
		}
		return next != null;
	}

	@Override
	public Entry<Point, T> next() {
		if (next != null) {
			Entry<Point, T> tmp = next;
			next = null;
			return tmp;
		}

		while (!stack.isEmpty()) {
			final QuadTree.Node<T> node = stack.pop();
			final Point pos = positions.pop();
			final int size = sizes.pop();

			if (node instanceof QuadTree.ContentNode || size == 1 << ignoreLevel) {
				return new Map.Entry<Point, T>() {
					@Override
					public Point getKey() {
						return pos;
					}

					@Override
					public T getValue() {
						if (node instanceof QuadTree.ContentNode) {
							return ((QuadTree.ContentNode<T>) node).content;
						}
						return null;
					}

					@Override
					public T setValue(T value) {
						throw new UnsupportedOperationException("Not supported yet.");
					}
				};
			} else if (size == 1) {
				throw new AssertionError("Only ContentNodes should be level 0");
			}

			add(node.se, new Point(pos.x + size / 2, pos.y + size / 2), size / 2);
			add(node.sw, new Point(pos.x, pos.y + size / 2), size / 2);
			add(node.ne, new Point(pos.x + size / 2, pos.y), size / 2);
			add(node.nw, new Point(pos.x, pos.y), size / 2);
		}
		return null;
	}

	@Override
	public void remove() {
		throw new UnsupportedOperationException("Not supported yet.");
	}
}
