package org.bitbucket.zetro.squaredino.io;

import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Map;
import java.util.Scanner;
import java.util.SortedMap;
import java.util.TreeMap;
import org.bitbucket.zetro.squaredino.backend.Board;
import org.bitbucket.zetro.squaredino.backend.BoardFactory;
import org.bitbucket.zetro.squaredino.backend.Cell;
import org.bitbucket.zetro.squaredino.misc.Point;

/**
 * Utility class for saving and loading cell boards.
 */
public class SquareIO {
	// Limit access
	private SquareIO() {
	}

	/**
	 * Writes a board and its data to a writer. Writes in RLE format.
	 *
	 * @param boardData the board that will be saved
	 * @param writer the writer saved to
	 * @throws IOException if an I/O error occurs
	 */
	public static void save(BoardData boardData, Writer writer) throws IOException {
		// #-lines
		if (boardData.getName() != null) {
			writer.write("#N " + boardData.getName() + "\n");
		}
		if (boardData.getAuthor() != null) {
			writer.write("#O " + boardData.getAuthor() + "\n");
		}
		if (boardData.getComments() != null) {
			for (String comment : boardData.getComments()) {
				writer.write("#C " + comment + "\n");
			}
		}

		// Header
		if (boardData.getWidth() != Board.INFINITE) {
			writer.write("x = " + boardData.getWidth() + ", ");
		}
		if (boardData.getHeight() != Board.INFINITE) {
			writer.write("y = " + boardData.getHeight() + ", ");
		}
		writer.write("rule = " + boardData.getRule());
		writer.write("\n");

		// Sorting cells
		int sx = Integer.MAX_VALUE;
		int sy = Integer.MAX_VALUE;
		SortedMap<Integer, SortedMap<Point, Integer>> s = new TreeMap<Integer, SortedMap<Point, Integer>>();
		for (Iterator<Map.Entry<Point, Cell>> it = boardData.getBoard().getLivingCells(); it.hasNext();) {
			Map.Entry<Point, Cell> entry = it.next();

			int y = entry.getKey().y;
			int x = entry.getKey().x;
			if (y < sy) {
				sy = y;
			}
			if (x < sx) {
				sx = x;
			}
			if (!s.containsKey(y)) {
				s.put(y, new TreeMap<Point, Integer>(new Comparator<Point>() {
					@Override
					public int compare(Point o1, Point o2) {
						return o1.x - o2.x;
					}
				}));
			}
			s.get(y).put(entry.getKey(), entry.getValue().state);
		}

		// Writing cells
		StringBuilder data = new StringBuilder();
		int y = sy;
		for (Iterator<Map.Entry<Integer, SortedMap<Point, Integer>>> it = s.entrySet().iterator(); it.hasNext();) {
			Map.Entry<Integer, SortedMap<Point, Integer>> entry = it.next();
			int dy = entry.getKey() - y;
			y = entry.getKey();
			if (dy > 0) {
				if (dy > 1) {
					data.append(dy);
				}
				data.append("$");
			}

			int x = sx - 1;
			int alive = 0;
			int cellState = 0;
			for (Iterator<Map.Entry<Point, Integer>> it2 = entry.getValue().entrySet().iterator(); it2.hasNext();) {
				Map.Entry<Point, Integer> entry2 = it2.next();
				Point point = entry2.getKey();
				int state = entry2.getValue();

				if (data.length() > 70) {
					writer.append(data + "\n");
					data = new StringBuilder();
				}

				int dx = point.x - x;
				x = point.x;
				if (dx > 1) {
					appendCell(data, cellState, alive);
					appendCell(data, 0, dx - 1);
					alive = 0;
				}
				if (cellState != state) {
					appendCell(data, cellState, alive);
					cellState = state;
					alive = 0;
				}
				alive++;
			}
			appendCell(data, cellState, alive);
		}
		data.append("!");
		writer.append(data);

		writer.close();
	}

	private static void appendCell(StringBuilder data, int state, int amount) {
		if (amount > 0) {
			if (amount > 1) {
				data.append(amount);
			}
			if (state == 0) {
				data.append("b");
			} else if (state == 1) {
				data.append("o");
			} else {
				int multiplier = (state - 1) / 24;
				int add = (state - 1) % 24;

				if (multiplier > 0) {
					data.append((char) (('p' - 1) + multiplier));
				}
				data.append((char) ('A' + add));
			}
		}
	}

	/**
	 * Loads a board and its data from a reader. Reads in RLE format.
	 *
	 * @param reader the reader loaded from
	 * @return the loaded board
	 * @throws IOException if an I/O error occurs
	 */
	public static BoardData load(Reader reader) throws IOException {
		Scanner scanner = new Scanner(reader);
		Board board = null;
		String name = null;
		String author = null;
		ArrayList<String> comments = new ArrayList<String>();
		int width = Board.INFINITE, height = Board.INFINITE;
		String rule = null;

		int startX = 0;
		int startY = 0;

		int xPos = 0;
		int yPos = 0;
		boolean parsedHeader = false;
		line:
		while (scanner.hasNext()) {
			String line = scanner.nextLine().trim();
			if (line.isEmpty()) {
				continue;
			}

			if (line.startsWith("#")) {
				if (parsedHeader) {
					System.err.println("Warning: #-line after header");
				}
				if (line.length() == 1) {
					continue;
				}
				char type = line.charAt(1);
				String nline = line.length() >= 2 ? line.substring(2).trim() : "";

				if (type == 'C' || type == 'c') {	// Comment (RLE)
					System.out.println("Comment: " + nline);
					comments.add(nline);
				} else if (type == 'D') {			// Description (MCell)
					System.out.println("Description: " + nline);
					name = nline;
				} else if (type == 'N') {			// Name (RLE)
					System.out.println("Name: " + nline);
					name = nline;
				} else if (type == 'O') {			// Author (RLE)
					System.out.println("Author: " + nline);
					author = nline;
				} else if (type == 'L') {			// Data line (MCell)
					System.out.println("Data line: " + nline);
					if (board == null) {
						board = BoardFactory.createBoard(width, height);
						if (rule != null) {
							board.setRule(rule);
						}
					}
					Point p = parseDataLine(board, nline, startX, xPos, yPos);
					xPos = p.x;
					yPos = p.y;
				} else if (line.startsWith("#RULE")) {	// Rule  (MCell)
					String r = line.substring(line.indexOf("RULE") + 4).trim();
					if (rule == null) {
						rule = r;
					} else {
						rule += r;
					}
					System.out.println("Rule: " + r);
				} else if (line.startsWith("#GAME")) {	// Game  (MCell)
					String r = line.substring(line.indexOf("GAME") + 4).trim();
					if (rule == null) {
						rule = r + ":";
					} else {
						rule = rule + ":" + r;
					}
					System.out.println("Game: " + r);
				} else {							// Unknown
					System.err.println("Warning: Unknown #-type '" + type + "' with content \"" + nline + "\"");
				}
			} else {
				if (!parsedHeader) {
					parsedHeader = true;

					System.out.println("Header: " + line);
					String[] s = line.split("\\s*,\\s*(?=.*=)");
					for (String string : s) {
						if (!string.contains("=")) {
							System.err.println("Error: Invalid header property \"" + string + "\"");
							continue;
						}

						String hname = string.substring(0, string.indexOf('=')).trim();
						String hvalue = string.substring(string.indexOf('=') + 1).trim();

						try {
							if (hname.equals("x")) {
								width = Integer.parseInt(hvalue);
								startX = -width / 2;
							} else if (hname.equals("y")) {
								height = Integer.parseInt(hvalue);
								startY = -height / 2;
							} else if (hname.equals("rule")) {
								rule = hvalue;
							} else {
								System.err.println("Warning: Unknown header type \"" + hname + " = " + hvalue + "\"");
							}
						} catch (NumberFormatException ex) {
							System.err.println("Error: \"" + hvalue + "\" isn't a valid number");
						}
					}

					board = BoardFactory.createBoard(width, height);
					if (rule != null) {
						board.setRule(rule);
					}

					xPos = startX;
					yPos = startY;
				} else {
					Point p = parseDataLine(board, line, startX, xPos, yPos);
					if (p == null) {
						break line;
					}
					xPos = p.x;
					yPos = p.y;
				}
			}
		}

		return new BoardData(board, name, author, comments, width, height);
	}

	private static Point parseDataLine(Board board, String line, int startX, int x, int y) {
		//System.out.println("Cell Line: " + line);
		int xPos = x, yPos = y;
		int runCount = 1;

		char[] cs = line.toCharArray();
		StringBuilder buf = new StringBuilder();
		for (int ii = 0; ii < cs.length; ii++) {
			char c = cs[ii];
			if (Character.isDigit(c)) {
				buf.append(c);
				continue;
			} else if (buf.length() > 0) {
				runCount = Integer.parseInt(buf.toString());
				buf.setLength(0);
			} else if (c >= 'p' && c <= 'z' && ii + 1 < cs.length) {
				char c2 = cs[ii + 1];
				if (c2 >= 'A' && c2 <= 'Z') {
					int state = (c - 'p' + 1) * 24 + (c2 - 'A' + 1);

					Cell cell = Cell.createCell(state);
					for (int i = 0; i < runCount; i++) {
						board.addCell(cell, xPos, yPos);
						xPos++;
					}
					runCount = 1;
					ii++;
					continue;
				}
			}
			if (c >= 'A' && c <= 'Z') {
				int state = c - 'A' + 1;

				Cell cell = Cell.createCell(state);
				for (int i = 0; i < runCount; i++) {
					board.addCell(cell, xPos, yPos);
					xPos++;
				}
				runCount = 1;
				continue;
			} else if (c == 'b' || c == '.') { // dead cell
				xPos += runCount;
				runCount = 1;
			} else if (c == 'o') { // alive cell
				Cell cell = Cell.createCell(1);
				for (int i = 0; i < runCount; i++) {
					board.addCell(cell, xPos, yPos);
					xPos++;
				}
				runCount = 1;
			} else if (c == '$') { // end of line
				xPos = startX;
				yPos += runCount;
				runCount = 1;
			} else if (c == '!') { // quit
				return null;
			} else {
				System.err.println("Warning: Unknown type '" + c + "' in \"" + line + "\"");
			}
		}

		return new Point(xPos, yPos);
	}
}
